#include "isonews_rdf.h"

////////////////////////////////////////////////////////////////
////              INITIALIZATION FUNCTIONS                  ////
////////////////////////////////////////////////////////////////

// This function is run when samurize first loads the plugin into memory
extern "C" __declspec(dllexport)
void __stdcall startup()
{
 	srand (time(NULL));
	hMutex = CreateMutex(0, false, "isonews_grabber");
}

// This function is run when samurize unloads the plugin from memory
extern "C" __declspec(dllexport)
void __stdcall cleanup()
{
	delete [] sInputUrl;
	delete [] sOutputPath;
	delete [] theFileName;
	delete [] releaseFormatString;
	delete [] stuffString;
	delete [] theFilePath;
	delete [] theURL;
	delete [] divxReturn;
	delete [] theReleaseType;
	delete [] theError;
	delete [] fileContents;
	delete [] replaceHelper;
}

BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			startup();
			break;
		case DLL_PROCESS_DETACH:
			cleanup();
			break;
	}
    return TRUE;
}

// This function is called by samurize to retrieve the active functions
extern "C" __declspec(dllexport)
char* __stdcall init()
{	
	return "isonewsALL|isonewsGames|isonewsCloneCD|isonewsUtil|isonewsKiddie|isonewsDOX|isonewsPS2|isonewsPS2DVD|isonewsXbox|isonewsDreamcast|isonewsGBA|isonewsTVRip|isonewsVCD|isonewsDVDRip|isonewsDVDR|isonewsDivX|isonewsAnime|isonewsAdult";
}

// This functions is called by samurize to get the parameters necessary for different functions
extern "C" __declspec(dllexport)
char* __stdcall getparam(char* func)
{
	return "how many releases|display format string|";
}

////////////////////////////////////////////////////////////////
////                 INTERNAL FUNCTIONS                     ////
////////////////////////////////////////////////////////////////


// helper function to GETINTERNETFILE() (does all the dirty work)
extern "C" __declspec(dllexport)
VOID __stdcall DownloadFile(HINTERNET hSource, HANDLE hDestination)
{
    UINT       uRetry;
    BOOL       fOk;
    BYTE       bFileBuffer [FILE_BUFFER_SIZE];

    dWriteFileTotal = 0;
    
    do
    {
        uRetry = 0;
        do
        {
            dReadFileCount = 0;
            fOk = InternetReadFile (hSource,
                                    bFileBuffer,
                                    FILE_BUFFER_SIZE,
                                    &dReadFileCount);
        }
        while (!(dReadFileCount || fOk || uRetry++ == RETRY_READ));

        if (dReadFileCount)
            {
            dWriteFileCount = 0;
            fOk = WriteFile (hDestination,
                             bFileBuffer,
                             dReadFileCount,
                             &dWriteFileCount,
                             NULL);
            if (fOk = fOk && (dReadFileCount == dWriteFileCount))
                dWriteFileTotal += dWriteFileCount;
            else dReadFileCount = 0;
            }
        }
    while (dReadFileCount);
}

// opens connection to internet to download a file
extern "C" __declspec(dllexport)
int __stdcall GetInternetFile(HINTERNET hSession, char sSource [], char sDestination [])
{
    HINTERNET hSource;
    HANDLE    hDestination;

    hSource = InternetOpenUrl (hSession,
                               sSource,
                               NULL,
                               0xFFFFFFFF,
                               INTERNET_FLAG_DONT_CACHE | 
                               INTERNET_FLAG_RAW_DATA,
                               0);
    if (hSource != NULL)
    {
        if (!sOutputPath [0]) hDestination = hStdOut;
        else
        {
            hDestination = CreateFile (sDestination,
                                       GENERIC_WRITE,
                                       FILE_SHARE_READ|FILE_SHARE_WRITE,
                                       NULL,
                                       CREATE_ALWAYS,
                                       FILE_FLAG_WRITE_THROUGH | 
                                       FILE_FLAG_SEQUENTIAL_SCAN,
                                       NULL);
        }
        if (hDestination != INVALID_HANDLE_VALUE)
        {
            DownloadFile (hSource, hDestination);

            if (hDestination != hStdOut) CloseHandle (hDestination);
        }
        else
		{
			InternetCloseHandle (hSource);
			strcpy(theError, "ERROR: Could not store downloaded data to file");
			return 1;
		}
    }
	else
	{
		sprintf(theError, "ERROR: specified url is currently unavailable\n%s", sSource);
		return 1;
	}
	return 0;
}

// initializes the connection to the internet, and initiates download ( calls GETINTERNETFILE() )
extern "C" __declspec(dllexport)
int __stdcall InternetSession(char sCaller [])
{
	int isError;
    HINTERNET hSession;

    hSession = InternetOpen (sCaller,
                             LOCAL_INTERNET_ACCESS,
                             NULL, 
                             INTERNET_INVALID_PORT_NUMBER,
                             INTERNET_FLAG_DONT_CACHE);
    if (hSession != NULL)
    {
        isError = GetInternetFile (hSession, sInputUrl, sOutputPath);
        InternetCloseHandle (hSession);
		return isError;
    }
	else
	{
		strcpy(theError, "ERROR: Unable to connect to Internet (failed to open session)");
		return 1;
	}
}

// copies arguments needed to get file from internet and send to file, and initiates internet connection
extern "C" __declspec(dllexport)
int HTTPFileDownload(char *filepath, char *httppath)
{
    sprintf (sInputUrl, "%s", httppath);
    sprintf (sOutputPath, "%s", filepath);
    return InternetSession ("blahblah");
}

// generates random number ( seeded in STARTUP() )
int  randNumGen()
{
	return (rand()%1000);
}

// locates temp folder, makes folder to store temp files in (if necessary), returns filename given LocationID and UnitType
char *getFileName()
{
	sprintf(theFilePath, "%s\\isonews.rss", getenv("TEMP"));
	if(!SetCurrentDirectory(theFilePath)) CreateDirectory(theFilePath, NULL);
	if(!strcmp("", theReleaseType)) sprintf(theFileName, "%s\\isonews-all.xml", theFilePath);
	else sprintf(theFileName, "%s\\isonews%s.xml", theFilePath, theReleaseType);
	return theFileName;
}

// helper function calculates the "age" of a file in minutes by checking the modified time of a file against the system time
int FileTimeDiff(LPFILETIME theold, LPFILETIME thenew)
{
	ULARGE_INTEGER uLargeIntegerTime1;
	ULARGE_INTEGER uLargeIntegerTime2;
	ULARGE_INTEGER uLargeIntegerTimeRESULT;
	FILETIME fileTimeRESULT;
	SYSTEMTIME systemTimeRESULT;

	uLargeIntegerTime1.LowPart = theold->dwLowDateTime;
    uLargeIntegerTime1.HighPart = theold->dwHighDateTime;
    uLargeIntegerTime2.LowPart = thenew->dwLowDateTime;
    uLargeIntegerTime2.HighPart = thenew->dwHighDateTime;
    uLargeIntegerTimeRESULT.HighPart = uLargeIntegerTime2.HighPart - uLargeIntegerTime1.HighPart;
    uLargeIntegerTimeRESULT.LowPart = uLargeIntegerTime2.LowPart  - uLargeIntegerTime1.LowPart;
    fileTimeRESULT.dwLowDateTime = uLargeIntegerTimeRESULT.LowPart;
    fileTimeRESULT.dwHighDateTime = uLargeIntegerTimeRESULT.HighPart;
	FileTimeToSystemTime( &fileTimeRESULT, &systemTimeRESULT);
	return systemTimeRESULT.wHour*60 + systemTimeRESULT.wMinute;
}

// checks to see if specified cache file needs to be refreshed ( file age > UPDATEINTERVAL )
int checkCache()
{
	HANDLE hFile=NULL;
	FILETIME mtime, ctime, atime, ztime;
	SYSTEMTIME stime;

	// open a handle to cached file
	hFile = CreateFile(getFileName(), GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return 1; // doesn't exist, so needs refreshing

	// get last modified time of file
	GetFileTime(hFile, &ctime, &atime, &mtime);
	CloseHandle(hFile);

	// get current system time
	GetSystemTime(&stime);
	SystemTimeToFileTime(&stime, &ztime);

	// compare modified time of file to system time
	// if older than specified interval, refresh the file
	if( FileTimeDiff(&mtime, &ztime) > UPDATEINTERVAL ) return 1;
	return 0;
}

// generates a URL given a releasetype, and tacks a random number to the end of it, stores URL in theURL
char *getURL()
{
	sprintf(theURL, "http://www.theisonews.com/rss/releases%s.rdf\?random=%d", theReleaseType, randNumGen());
	return theURL;
}

// string replacement helper function
char *strreplace(char *source, char *theOld, char *theNew)
{	
	char *index;
	replaceHelper[0] = '\0';
	index = strstr(source, theOld);
	if(index)
	{
		strncpy(replaceHelper, source, (int)(index - source));
		replaceHelper[index-source]='\0';
		strcat(replaceHelper, theNew);
		strcat(replaceHelper, index+strlen(theOld));
		strcpy(source, replaceHelper);
	}
	if(!index) return source;
	else return source;
}

// reads file contents to buffer
void readFileBegin(char *fname)//Get chunk from beginning of file
{
	FILE *inFile;
	static unsigned int fileSize;

	inFile = fopen(fname, "rb");
	fseek (inFile , 0 , SEEK_END);
	fileSize = ftell (inFile);
	rewind (inFile);
	if(strlen(fileContents)>MAXFILESIZE) fileSize = MAXFILESIZE;
	fread(fileContents, 1, fileSize, inFile);
	fclose(inFile);
}

// parses the rdf file
int getReleaseNum(int releaseNum)
{
	char *thePos;
	int howFar;
	thePos = fileContents;
	for(howFar=0; howFar<(releaseNum+2); howFar++) thePos = strstr(thePos, "<title>") + 7;
	returnVal = thePos;
	thePos = strstr(returnVal, "</title>");
	return thePos - returnVal;
}

// called by all user functions, checks cached file, and refreshes it if necessary, then parses the XML
char *getData(char *ReleaseType)
{
	strcpy(theReleaseType, ReleaseType); // copy arguments into memory (compatibility)
	if(checkCache())
	{
		if(HTTPFileDownload(getFileName(), getURL())) return theError;
	}
	return "no_error";
}

// parses extract from rdf file, into user defined chunks
char *extractStuff(char *theData, int which)
{
	static char *holder, *endholder;
	static int x;
	holder = theData;
	for(x=0; x<which; x++)
	{
		holder = strstr(holder, "^")+1;
		endholder = strstr(holder, "^");
	}
	if(which == 4) strcpy(stuffString, holder);
	else
	{
		strncpy(stuffString, holder, endholder - holder);
		stuffString[endholder - holder] = '\0';
	}
	return stuffString;
}

// does the custom format string display stuff
int formatRelease(char *formatString, char *theData)
{
	strreplace(theData, " released by ", "^");
	strreplace(theData, " (", "^");
	strreplace(theData, ", ", "^");
	strreplace(theData, ")", "");

	if(!((int)strstr(formatString, "<title>") + (int)strstr(formatString, "<group>") + (int)strstr(formatString, "<date>") + (int)strstr(formatString, "<size>")))
	{
		strcpy(divxReturn, "use <title>, <group>, <date>, <size> in your format string\nIE, <title> - <group> [<size>]");
		return 1;
	}

	strreplace(formatString, "<title>", extractStuff(theData, 1));
	strreplace(formatString, "<group>", extractStuff(theData, 2));
	strreplace(formatString, "<date>", extractStuff(theData, 3));
	strreplace(formatString, "<size>", extractStuff(theData, 4));
	return 0;
}

// main plugin function
char *getSection(char *section, char *howMany, char *formatString)
{
	static int thereleasenum;
	if (WaitForSingleObject(hMutex,5000) == WAIT_TIMEOUT)
	{
		ReleaseMutex(hMutex);
		return "ERROR: lock timeout";
	}

	if(strcmp(getData(section), "no_error"))
	{
		ReleaseMutex(hMutex);
		return theError; // refresh data (if necessary)
	}

	readFileBegin(getFileName());
	divxReturn[0] = '\0';

	for(thereleasenum = 1; thereleasenum<=atoi(howMany); thereleasenum++)
	{
		strncpy(theError+1, returnVal, getReleaseNum(thereleasenum));
		theError[0] = '^';
		theError[getReleaseNum(thereleasenum)] = '\0';
		strcpy(releaseFormatString, formatString);
		if(formatRelease(releaseFormatString, theError)) break;
		if(thereleasenum!=1) sprintf(divxReturn, "%s\n%s", divxReturn, releaseFormatString);
		else sprintf(divxReturn, "%s", releaseFormatString);
	}

	ReleaseMutex(hMutex);
	return divxReturn;
}

////////////////////////////////////////////////////////////////
////               FUNCTIONS SHOWN TO USER                  ////
////////////////////////////////////////////////////////////////

extern "C" __declspec(dllexport)
char* __stdcall isonewsALL(char *howMany, char *formatString)
{
	return getSection("", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsGames(char *howMany, char *formatString)
{
	return getSection("-game", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsCloneCD(char *howMany, char *formatString)
{
	return getSection("-clonecd", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsUtil(char *howMany, char *formatString)
{
	return getSection("-util", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsKiddie(char *howMany, char *formatString)
{
	return getSection("-kiddie", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsDOX(char *howMany, char *formatString)
{
	return getSection("-pd", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsPS2(char *howMany, char *formatString)
{
	return getSection("-ps2", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsPS2DVD(char *howMany, char *formatString)
{
	return getSection("-ps2dvd", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsXbox(char *howMany, char *formatString)
{
	return getSection("-xbox", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsDreamcast(char *howMany, char *formatString)
{
	return getSection("-dc", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsGBA(char *howMany, char *formatString)
{
	return getSection("-gba", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsTVRip(char *howMany, char *formatString)
{
	return getSection("-tvrip", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsVCD(char *howMany, char *formatString)
{
	return getSection("-vcd", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsDVDRip(char *howMany, char *formatString)
{
	return getSection("-dvdrip", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsDVDR(char *howMany, char *formatString)
{
	return getSection("-dvdr", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsDivX(char *howMany, char *formatString)
{
	return getSection("-divx", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsAnime(char *howMany, char *formatString)
{
	return getSection("-anime", howMany, formatString);
}

extern "C" __declspec(dllexport)
char* __stdcall isonewsAdult(char *howMany, char *formatString)
{
	return getSection("-adult", howMany, formatString);
}